from ase import io
from xespresso import Espresso
import numpy as np
import copy
import os
import shutil
import sys


class HpXEspresso(Espresso):
    def __init__(self, restart=None, ignore_bad_restart_file=False,
                 label='xespresso', atoms=None, package='pw', parallel='',
                 queue=None, queue_hp=False, parallel_hp='', iteration=1,
                 sc_type="relax", relax_type="vc-relax",
                 insulator=None, magnetic=None, nbnd=None,
                 parallelization_hp=False, ethr_relax=None, ethr_scf=None, ethr_scf2=None,
                 **kwargs):
        """
        This class allows to perform the chain of relax-scf-hp calculations necessary to compute U/V values
        via DFPT "Hubbard parameters from density-functional perturbation theory", 
        Phys. Rev. B 98, 085127 (2018); arXiv:1805.01805 using ase and XEspresso, 
        of which XpXEspresso is a subclass.

        input_data, pseudopotentials, kspacing, kpts, koffset
            Please have a look at Espresso module in ASE

        queue: dict
            A dictionary with parameters for job submission of the pw.x calculation, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}

        queue_hp: dict
            A dictionary with parameters for job submission of the hp.x calculation, e.g.
             queue = {'nodes': 4, 'ntasks-per-node': 20, 
                      'account': 'xxx', 'partition': 'normal', 
                      'time': '23:59:00'}
        package: str
            Choose the quantum espresso pacakge: pw, dos, projwfc, band, pp, ph, ..
            For NEB calculation, please use neb.NEBEspresso module.
            Calculaiton use phonon is not implemented yet.

        parallel: str
            A str which control the parallelization parameters for the pw.x runs: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).

        parallel_hp: str
            A str which control the parallelization parameters for the hp.x runs: -nimage, -npools, 
            -nband, -ntg, -ndiag or -northo (shorthands, respectively: 
            -ni, -nk, -nb, -nt, -nd).

        max_iter: int
            An int indicating the interation step in case a self-consistent (site-dependent) calculation
            of the U/V parameters is performed (Chech the Uscsd module).

        sc_type: str 
            A str indicating if the procedure for the self-consistent (site-dependent) calculation
            of the U/V parameters is performed by relaxing or not the structure. Check also the Uscsd class.
            Possible values are "relax" and "scf".

        relax_type: str
            Possible values are "vc-relax" and "relax". Indicate thetype of relaxation run incase sc_type = "relax".
            Check also the Uscsd class.

        insulator: bool
            A bool indicating if the material is an insulator (True) or a metal (False).
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.

        magnetic: int
            An int indicating if the material is non-magneticc (1) or  magnetic (2)
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.

        nbnd: int
            An int indicating the number of bands to take into account in the calculations.
            It has to be specified when sc_type = "scf" and iteration = 1, otherwhise one could not determine 
            the correct procedure to be perfomed.

        parallelization_hp: bool
            A bool indicating if  parallelization of the hp run over atoms or qpoints needs to be performed.
            parallelization_hp = True

        ethr_relax: float
            A float corresponfing to the threshold for the scf convergence in the relaxation step.  
            If not specified it is set to 1.0E-8

        ethr_scf: float
            A float corresponfing to the threshold for the scf convergence in the (first) scf step.
            If not specified it is set to 1.0E-10

        ethr_scf2: float
            A float corresponfing to the threshold for the scf convergence in the second scf step.
            If not specified it is set to 1.0E-12

        !!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!
        Do not forget to provide in the input_data dictionry and two elements:
            "lda_plus_U" : True
            "input_ntyp": { "starting_magnetization" : {Hubbard_site: XX},
                            "Hubbard_U" : {Hubbard_site: XX}}

        TO DO:
        - check inputs: sc_type, relax_typ, insulator, magnetic, ndnb
        - implement parallelization_hp = True
        - implement possibility of restart of an interrupted relax run or hp run.
        - check copy parameters.in file in the case of DFT+U+V

        """
        super().__init__(restart=None, ignore_bad_restart_file=False,
                         label='xespresso', atoms=None, package='pw', parallel='',
                         queue=None,
                         **kwargs)

        self.calc = None

        self.relax_directory = None
        self.relax_parameters = None
        self.relax_results = None

        self.scf_directory = None
        self.scf_parameters = None
        self.scf_results = None

        self.hp_directory = None
        if 'parameters_hp' not in kwargs:
            kwargs['parameters_hp'] = {}
        self.hp_parameters = kwargs['parameters_hp']
        self.hp_results = None

        self.directory = os.path.split(label)[0]
        self.prefix = os.path.split(label)[1]
        self.directory_ref = self.directory

        self.queue_hp = queue_hp
        self.parallel_hp = parallel_hp
        self.queue = queue
        self.parallel = parallel

        self.iteration = iteration

        self.sc_type = sc_type
        self.relax_type = relax_type

        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a str for the sc_type variable: relax or scf")

        if self.relax_type != None and self.relax_type not in ["relax", "vc-relax"]:
            sys.exit(
                "Please provide a str for the relax_type variable: relax or vc-relax")
        if self.sc_type != None and self.sc_type not in ["relax", "scf"]:
            sys.exit("Please provide a bool for the sc_type variable: relax or scf")

        self.insulator = insulator
        self.magnetic = magnetic
        self.nbnd = nbnd

        if self.insulator != None and bool(self.insulator) == False:
            sys.exit("Please provide a bool for the insulator variable")

        if self.magnetic != None and self.magnetic not in [1, 2]:
            sys.exit(
                "Please provide an int for the magnetic  variable: 1 for spin unpolirzed and 2 for spin poolariz")

        if self.nbnd != None and type(self.nbnd) != int:
            sys.exit("Please provide an int for the nbnd variable")

        self.parallelization_hp = parallelization_hp

        self.ethr_relax = ethr_relax
        self.ethr_scf = ethr_scf
        self.ethr_scf2 = ethr_scf2

    def read_results_fromdirectory(self, directory):
        """
        Extension of the read_results metho of the XEspresso class, so that results
        can be read from a directory different from self.directory

        directory: str
            A str indicating the the path of the directory from which results have to be read.
        """
        from xespresso.xio import get_atomic_species
        pwo = self.prefix + '.pwo'

        if directory == None:
            directory = self.directory

        pwos = [file for file in os.listdir(directory) if pwo in file]
        output = None
        for pwo in pwos:
            atomic_species = None
            pwo = os.path.join(directory, pwo)
            # print('read results from: %s'%pwo)
            # output = next(read_espresso_out(pwo))
            atomic_species = get_atomic_species(pwo)
            try:
                output = io.read(pwo)
                atomic_species = get_atomic_species(pwo)
                if atomic_species:
                    output.arrays['species'] = atomic_species
            except Exception as e:
                print(pwo, e)
                # print('\nread %s failed\n'%pwo)
        if output:
            self.calc = output.calc
            self.results = output.calc.results
            self.results['atoms'] = output
            self.efermi = self.get_fermi_level()
            self.nspins = self.get_number_of_spins()
            self.atoms = output
        else:
            print('\nRead result failed\n')

    def is_insulator(self, fermi=None):
        """
        Checking if the system is a insulators
        One can provide the fermi energy as input, otherwise it will be extracted from the pw.pwo file.
        WARNING: The Fermi energy cannot beread from a calculation run using "tot_magnetization"

        fermi: float
            A float corresponding to the Fermi energy

        Results:
        - set with Bool (True=insulator, False=metal), bandgap value (eV), VBM and CBm values (eV)
        """

        if fermi == None:
            fermi = self.calc.get_fermi_level()
            if fermi == None:
                raise Exception("I cannot get the Fermi energy from the pw.pwo file"
                                "Provide a Fermi energy value.")

        # reorganize the bands, rather than per kpoint, per energy level
        bands = []
        nbands = len(self.calc.get_eigenvalues(kpt=0, spin=0))
        for i in range(nbands):
            bands.append([])
            for k in range(len(self.calc.get_k_point_weights())):
                if self.calc.get_number_of_spins() == 1:  # calc.nspins == 1:
                    bands[i].append(self.calc.get_eigenvalues(kpt=k)[i])
                else:
                    bands[i].append(
                        self.calc.get_eigenvalues(kpt=k, spin=0)[i])
                    bands[i].append(
                        self.calc.get_eigenvalues(kpt=k, spin=1)[i])
        bands = np.array(bands)
        max_mins = [(max(i), min(i)) for i in bands]

        # one band is crossed by the fermi energy
        if any(i[1] < fermi and fermi < i[0] for i in max_mins):
            return(False, None, None, None)

        # case of semimetals, fermi energy at the crossing of two bands
        # this will only work if the dirac point is computed!
        elif (any(i[0] == fermi for i in max_mins) and
              any(i[1] == fermi for i in max_mins)):
            return False, 0., fermi, fermi
        # insulating case
        else:
            # take the max of the band maxima below the fermi energy
            homo = max([i[0] for i in max_mins if i[0] < fermi])
            # take the min of the band minima above the fermi energy
            lumo = min([i[1] for i in max_mins if i[1] > fermi])
            gap = lumo - homo
            if gap <= 0.:
                raise Exception("Something wrong has been implemented. "
                                "Revise the code!")
            return True, gap, homo, lumo

    def get_magnetizations(self, getall=False):
        """
        Read the total and absolute magnetizations from the pw.x out file
        """

        start = '!    total energy'
        end = '     convergence has'

        mags = dict()
        mags['total magnetization'] = []
        mags['absolute magnetization'] = []

        filename = os.path.join(self.directory, self.prefix + ".pwo")
        with open(filename, 'r') as fobj:
            for line in fobj:
                if start in line:
                    while end not in line:
                        line = next(fobj)
                        for mag in mags.keys():
                            if mag in line:
                                mags[mag].append(float(line.split()[3]))
        if getall:
            return mags
        else:
            return [mags[k][-1] if mags[k] else 0.0 for k in
                    sorted(mags.keys())]

    def read_Hubbard_parameters(self, iteration):
        """
        Read the U/V values from the Hubbard_parameters.dat file or from the paramets.in file for DFT+U+V
        """

        if 'lda_plus_u_kind' not in self.parameters or self.parameters['lda_plus_u_kind'] != 1:
            start = 'site n.'
            hp_U = dict()
            directory = os.path.join(self.directory_ref, "hp_"+str(iteration))
            filename = os.path.join(
                directory, self.prefix + ".Hubbard_parameters.dat")

            print(filename)
            try:
                file = open(filename, 'r')
            except IOError:
                print('There was an error opening the Hubbard.dat file!')
                sys.exit('There was an error opening the Hubbard.dat file!')

            with open(filename, 'r') as fobj:
                lines = fobj.readlines()
                nlines = len(lines)
                start = [n for n, line in enumerate(
                    lines) if 'site n.' in line][0]
                end = [n for n, line in enumerate(
                    lines) if "=--------------" in line][1] - 1
                for line in lines[start+1:end]:
                    l = line.split()
                    hp_U[l[0]] = {}
                    hp_U[l[0]]['label'] = l[2]
                    hp_U[l[0]]['spin'] = l[3]
                    hp_U[l[0]]['type'] = l[4]
                    hp_U[l[0]]['new_label'] = l[5]
                    hp_U[l[0]]['U'] = l[6]

        # elif self.scf_parameters['lda_plus_u_kind'] == 2:
        #    hp_U = None
        #    print("Reading the parametes.in file for DFT+U+V is not implemented yet")
        else:
            hp_U = None
            print("lda_plus_u_kind=1 is not implemented")

        return hp_U

    def run_relax(self):
        """
        Running the relaxation step
        """

        if self.sc_type == "scf":
            sys.exit("You set sc_type to False. I will not relax the structure")

        self.directory = os.path.join(
            self.directory_ref, "relax_"+str(self.iteration))
        self.set_queue(package="pw", parallel=self.parallel, queue=self.queue)

        if self.relax_parameters != None:  # This works if the execution of the script is not interrupted
            self.parameters = copy.deepcopy(self.relax_parameters)

        self.parameters["calculation"] = self.relax_type
        self.parameters['verbosity'] = 'high'
        if self.ethr_relax == None:
            self.parameters["conv_thr"] = 1.0E-8
        else:
            self.parameters["conv_thr"] = self.ethr_relax

        self.parameters['input_data']['restart_mode'] = 'from_scratch'

        if self.iteration > 1:
            if 'lda_plus_u_kind' in self.parameters['input_data'] and self.parameters['input_data']['lda_plus_u_kind'] == 2:
                old_parametersin = os.path.join(
                    self.directory_ref, "hp_"+str(self.iteration - 1), 'parameters.in')
                shutil.copy(old_parametersin,
                            self.directory, symlinks=False, ignore=None)
                self.parameters['input_data']['Hubbard_parameters'] = 'file'

    def run_scf(self):
        """
        Running the (first) scf step 
        """
        #self.set_queue(package = "pw", parallel = self.parallel, queue = self.queue)
        if self.sc_type == "relax":
            # self.read_results()
            if self.iteration == 1:
                self.read_results()
            else:
                previous_run = os.path.join(
                    self.directory_ref, "relax_"+str(self.iteration))
                self.read_results_fromdirectory(directory=previous_run)
                self.relax_directory = previous_run

            if not self.relax_directory:
                self.relax_directory = self.directory
            if not self.relax_parameters:
                self.relax_parameters = copy.deepcopy(self.parameters)
            if not self.relax_results:
                self.relax_results = copy.deepcopy(self.results)
                self.efermi = self.get_fermi_level()
            # read new atoms from results and set magmoms
            self.atoms = self.results['atoms']
            self.parameters = copy.deepcopy(self.relax_parameters)
            self.parameters['input_data']['restart_mode'] = 'restart'

            # create working directory
            self.save_directory_old = os.path.join(
                self.relax_directory, '%s.save' % self.prefix)
            self.directory = os.path.join(
                self.directory_ref, "scf_"+str(self.iteration))
            self.save_directory = os.path.join(
                self.directory, '%s.save' % self.prefix)
            self.label = os.path.join(self.directory, self.prefix)
            self.scf_directory = self.directory

            if self.iteration > 1:
                if 'lda_plus_u_kind' in self.parameters['input_data'] and self.parameters['input_data']['lda_plus_u_kind'] == 2:
                    old_parametersin = os.path.join(
                        self.directory_ref, "hp_"+str(self.iteration - 1), 'parameters.in')
                    shutil.copy(old_parametersin,
                                self.directory, symlinks=False, ignore=None)
                    self.parameters['input_data']['Hubbard_parameters'] = 'file'

            # if path already exists, remove it before copying with copytree()
            if os.path.exists(self.save_directory):
                shutil.rmtree(self.save_directory)

            shutil.copytree(self.save_directory_old,
                            self.save_directory, symlinks=False, ignore=None)
            # self.calculate()
            insulator = self.is_insulator()[0]
            magnetic = self.get_number_of_spins()
            nbnd = len(self.calc.get_eigenvalues(kpt=0))
        else:
            self.directory = os.path.join(
                self.directory_ref, "scf_"+str(self.iteration))
            self.scf_directory = self.directory

            if self.iteration == 1:
                if self.insulator == None:
                    sys.exit(
                        "You need to provide a value (True/False) for  'insulator'")
                else:
                    insulator = self.insulator
                if self.magnetic == None:
                    sys.exit(
                        "You need to provide a value (True/False) for  'magnetic'")
                else:
                    magnetic = self.magnetic
                if self.nbnd == None:
                    sys.exit(
                        "You need to provide a value (True/False) for  'nbnd'")
                else:
                    nbnd = self.nbnd

            else:
                previous_run = os.path.join(
                    self.directory_ref, "scf_"+str(self.iteration-1))
                self.read_results_fromdirectory(directory=previous_run)
                insulator = self.is_insulator()[0]
                magnetic = self.get_number_of_spins()
                nbnd = len(self.calc.get_eigenvalues(kpt=0))

        self.parameters['verbosity'] = 'high'
        self.parameters['input_data']['verbosity'] = 'high'
        self.parameters['calculation'] = "scf"

        if insulator == True and magnetic == 1:
            self.parameters['input_data']["occupations"] = "fixed"
            #nbnd = len(self.calc.get_eigenvalues(kpt=0))
            self.parameters['input_data']["nbnd"] = nbnd
            if "degauss" in self.parameters['input_data'].keys():
                del self.parameters['input_data']["degauss"]

            if "starting_magnetization" in self.parameters['input_data']['input_ntyp'].keys():
                del self.parameters['input_data']['input_ntyp']["starting_magnetization"]

            if self.ethr_scf == None:
                self.parameters["conv_thr"] = 1.0E-10
            else:
                self.parameters["conv_thr"] = self.ethr_scf

        elif insulator == False or (insulator == True and magnetic == 2):
            self.parameters['input_data']["occupations"] = "smearing"
            self.parameters['input_data']["degauss"] = 0.01

            if self.ethr_scf == None:
                self.parameters["conv_thr"] = 1.0E-10
            else:
                self.parameters["conv_thr"] = self.ethr_scf

        else:
            if self.sc_type == "relax":
                sys.exit("Case not implemented: is an insulator? {} and is spin polarized? {}".format(
                    self.is_insulator()[0], self.get_spin_polarized()))
            else:
                sys.exit("Case not implemented: is an insulator? {} and is spin polarized? {}".format(
                    self.insulator, self.magnetic))

        self.set_queue(package="pw", parallel=self.parallel, queue=self.queue)

    def run_scf2(self):
        """
        Running the second scf step (if necessary)
        """
        self.read_results()
        if self.get_number_of_spins() == 2 and self.is_insulator()[0] == True:

            # save relax parameters
            if not self.scf_directory:
                self.scf_directory = self.directory
            if not self.scf_parameters:
                self.scf_parameters = copy.deepcopy(self.parameters)
            if not self.scf_results:
                self.scf_results = copy.deepcopy(self.results)
                self.efermi = self.get_fermi_level()

            # read new atoms from results and set scf specific parameters
            self.atoms = self.results['atoms']
            self.parameters = copy.deepcopy(self.scf_parameters)
            self.parameters['calculation'] = "scf"
            self.parameters['input_data']['verbosity'] = 'high'
            self.parameters['input_data']['restart_mode'] = 'restart'
            self.parameters['input_data']["occupations"] = "fixed"
            nbnd = len(self.calc.get_eigenvalues(kpt=0))
            self.parameters['input_data']["nbnd"] = nbnd
            self.parameters['input_data']["tot_magnetization"] = abs(
                self.get_magnetizations(True)['total magnetization'][-1])

            if "degauss" in self.parameters['input_data'].keys():
                del self.parameters['input_data']["degauss"]

            if "starting_magnetization" in self.parameters['input_data']['input_ntyp'].keys():
                del self.parameters['input_data']['input_ntyp']["starting_magnetization"]

            self.parameters['input_data']["startingpot"] = "file"
            self.parameters['input_data']["startingwfc"] = "file"

            if self.ethr_scf2 == None:
                self.parameters["conv_thr"] = 1.0E-12
            else:
                self.parameters["conv_thr"] = self.ethr_scf2

            # create working directory
            self.save_directory_old = os.path.join(
                self.scf_directory, '%s.save' % self.prefix)
            self.directory = os.path.join(
                self.directory_ref, "scf2_"+str(self.iteration))
            self.save_directory = os.path.join(
                self.directory, '%s.save' % self.prefix)
            self.label = os.path.join(self.directory, self.prefix)
            self.scf_directory = self.directory

            # if path already exists, remove it before copying with copytree()
            if os.path.exists(self.save_directory):
                shutil.rmtree(self.save_directory)

            shutil.copytree(self.save_directory_old,
                            self.save_directory, symlinks=False, ignore=None)

            self.scf2_necessary = True
        else:
            self.scf2_necessary = False
            print("No need to perform a second scf")
            pass

        self.set_queue(package="pw", parallel=self.parallel, queue=self.queue)

    def run_hp(self):
        """
        Running the hp step
        """
        # create working directory
        if self.scf_directory == None:
            if os.path.exists(os.path.join(self.directory_ref, "scf2_"+str(self.iteration))):
                self.scf_directory = os.path.join(
                    self.directory_ref, "scf2_"+str(self.iteration))
            else:
                self.scf_directory = os.path.join(
                    self.directory_ref, "scf_"+str(self.iteration))

        if self.iteration == 1:
            self.read_results()
        else:
            self.read_results_fromdirectory(directory=self.scf_directory)

        if self.parallelization_hp == False:
            if not self.hp_parameters:
                self.hp_parameters = copy.deepcopy(self.parameters_hp)
            if not self.hp_results:
                self.hp_results = copy.deepcopy(self.results)

            self.save_directory_old = os.path.join(
                self.scf_directory, '%s.save' % self.prefix)
            self.directory = os.path.join(
                self.directory_ref, "hp_"+str(self.iteration))
            #self.directory = os.path.join(self.directory_ref,"hp_",str(iteration))
            self.save_directory = os.path.join(
                self.directory, '%s.save' % self.prefix)
            self.label = os.path.join(self.directory, self.prefix)
            self.hp_directory = self.directory

            # if path already exists, remove it before copying with copytree()
            if os.path.exists(self.save_directory):
                shutil.rmtree(self.save_directory)

            shutil.copytree(self.save_directory_old,
                            self.save_directory, symlinks=False, ignore=None)

            self.post(package='hp', queue=self.queue_hp,
                      parallel=self.parallel_hp, **self.hp_parameters)
        else:
            sys.exit(
                "Parallelization of the hp.x calculation over the atoms or q-points not implemented.")



import numpy as np


def get_hash(file):
    import hashlib
    with open(file, "rb") as f:
        file_hash = hashlib.md5()
        chunk = f.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(8192)
    hd = file_hash.hexdigest()
    return hd


def check_type(key, value, document):
    """
    """
    for section, parameters in document.items():
        if key in parameters:
            parameter_type = document[section][key][0]
            # print(type(value))
            if parameter_type == "CHARACTER":
                assert isinstance(
                    value, str), '\n  Parameter: %s, should be a string!' % key
            elif parameter_type == 'REAL':
                assert isinstance(value, float) or isinstance(
                    value, int), '\n  Parameter: %s, should be a float!' % key
            elif parameter_type == 'INTEGER':
                assert isinstance(
                    value, int), '\n  Parameter: %s, should be a integer!' % key
            elif parameter_type == 'LOGICAL':
                assert isinstance(
                    value, bool), '\n  Parameter: %s, should be a bool!' % key
            option_list = document[section][key][1]
            if len(option_list) > 0:
                assert value in option_list, '\n  Parameter: %s = %s, should be in %s!' % (
                    key, value, str(option_list))
            return
    assert '\n  Parameter: %s, is not a PWSCF parameter!' % key


def modify_text(text, type='CHARACTER'):
    if text == None:
        text = ''
    else:
        text = str(text)
    text = text.replace(" ", "")
    text = text.replace("'", "").replace(":", "")
    text = text.strip()
    if type == 'LOGICAL' and text.upper() in '.TRUE.':
        text = True
    elif type == 'LOGICAL' and text.upper() in '.FALSE.':
        text = False
    if type.upper() == 'CHARACTER':
        if not text:
            text = ''
    if type.upper() == 'REAL':
        text = text.replace("D", "E")
        try:
            text = float(text)
        except:
            print('Failed: text %s' % text)
    if type.upper() == 'INTEGER':
        try:
            text = int(text)
        except:
            pass
    return text



import os
import logging
logger = logging.getLogger(__name__)


config_files = [os.path.join(os.environ['HOME'], '.xespressorc'),
                '.xespressorc']


def set_queue(calc, package=None, parallel=None, queue=None, command=None):
    '''
    If queue, change command to "sbatch .job_file".
    The queue information are written into '.job_file'
    '''
    if queue is None:
        queue = calc.queue
    else:
        calc.queue = queue
    if package is None:
        package = calc.package
    else:
        calc.package = package
    if parallel is None:
        parallel = calc.parallel
    else:
        calc.parallel = parallel
    if command is None:
        command = os.environ.get('ASE_ESPRESSO_COMMAND')
    if 'PACKAGE' in command:
        if 'pw' in package:
            command = command.replace('PACKAGE', package, 1)
            command = command.replace('PACKAGE', 'pw', 2)
        else:
            command = command.replace('PACKAGE', package)
    if 'PREFIX' in command:
        command = command.replace('PREFIX', calc.prefix)
    if 'PARALLEL' in command:
        command = command.replace('PARALLEL', parallel)
    logger.debug('Espresso command: %s' % (command))
    if queue:
        script = ''
        if 'config' in queue:
            cf = os.path.join(os.environ['HOME'], queue['config'])
            file = open(cf, 'r')
            script = file.read()
            # del queue['config']
        else:
            for cf in config_files:
                if os.path.exists(cf):
                    file = open(cf, 'r')
                    script = file.read()
        jobname = calc.prefix
        with open('%s/.job_file' % calc.directory, 'w') as fh:
            fh.writelines("#!/bin/bash\n")
            fh.writelines("#SBATCH --job-name=%s \n" % jobname)
            fh.writelines("#SBATCH --output=%s.out\n" % calc.prefix)
            fh.writelines("#SBATCH --error=%s.err\n" % calc.prefix)
            fh.writelines("#SBATCH --wait\n")
            for key, value in queue.items():
                if key == 'config':
                    continue
                if value:
                    fh.writelines("#SBATCH --%s=%s\n" % (key, value))
            fh.writelines("%s \n" % script)
            fh.writelines("%s \n" % command)
        calc.command = "sbatch {0}".format('.job_file')
    else:
        calc.command = command
    logger.debug('Queue command: %s' % (calc.command))



"""
Creating a json file from the input_pw.html file
"""

import pprint
import requests
from bs4 import BeautifulSoup
from requests.api import options
import json
# from xespresso.utils import modify_text


def modify_text(text, type='CHARACTER'):
    if text == None:
        text = ''
    else:
        text = str(text)
    text = text.replace(" ", "")
    text = text.replace("'", "").replace(":", "")
    text = text.strip()
    if type == 'LOGICAL' and text.upper() in '.TRUE.':
        text = True
    elif type == 'LOGICAL' and text.upper() in '.FALSE.':
        text = False
    if type.upper() == 'CHARACTER':
        if not text:
            text = ''
    if type.upper() == 'REAL':
        text = text.replace("D", "E")
        try:
            text = float(text)
        except:
            print('Failed:   %s    %s' % (text, type))
            text = 1e30
    if type.upper() == 'INTEGER':
        try:
            text = int(text)
        except:
            print('Failed:   %s    %s' % (text, type))
            text = 1e30
    return text


packages = ['PW', 'PP', 'NEB', 'BANDS', 'DOS', 'PROJWFC']
parameters = {}
defaults = {}
for package in packages:
    print("{0:=^60}".format(package))
    parameters[package] = {}
    defaults[package] = {}
    url = 'https://www.quantum-espresso.org/Doc/INPUT_%s.html' % package
    html_text = requests.get(url).text
    soup = BeautifulSoup(html_text, 'html.parser')
    sections = soup.find_all('table', attrs={"border": "0"})
    for section in sections:
        section_header = section.find_all('h2')[0].text
        section_type = section_header.split(':')[0].strip()
        section_name = section_header.split(':')[1].strip().split()[0]
        print(section_type, section_name)
        tables = section.find_all('table', attrs={'width': "100%"})
        parameters[package][section_name] = {}
        defaults[package][section_name] = {}
        for table in tables:
            trs = table.find_all('tr')
            key = trs[0].find_all('th')[0].text.strip()
            if len(trs[0].find_all('th')) == 2:
                type = trs[0].find_all('th')[0].text.strip()
            else:
                type = trs[0].find_all('td')[0].text.strip()
            if len(trs) == 3:
                default = modify_text(trs[1].find_all('td')[1].text, type=type)
                options = trs[2].find_all('tt')
                option_list = []
                for option in options:
                    for text in option.text.split(','):
                        value = [modify_text(text, type=type)]
                        option_list = option_list + value
            else:
                default = modify_text('', type=type)
                options = trs[1].find_all('tt')
                option_list = []
                for option in options:
                    value = modify_text(option.text, type=type).split(',')
                    option_list = option_list + value
            parameters[package][section_name][key] = [type, option_list]
            defaults[package][section_name][key] = default

pprint.pprint(parameters)
pprint.pprint(defaults)
with open('input_parameters.json', 'w') as f:
    json.dump(parameters, f, indent=2)
with open('input_parameters.json', 'w') as f:
    json.dump(default, f, indent=2)




from xespresso.post.base import PostCalculation


class EspressoMatdyn(PostCalculation):

    package = 'matdyn'
    package_parameters = {'INPUT': ['flfrc', 'asr', 'dos', 'nk1', 'nk2', 'nk3', 'deltaE', 'ndos', 'fldos',
                                    'flfrq', 'flvec', 'fleig', 'fldvn', 'at', 'l1', 'l2', 'l3', 'ntyp',
                                    'amass', 'readtau', 'fltau', 'la2F', 'q_in_band_form', 'q_in_cryst_coord',
                                    'eigen_similarity', 'fd', 'na_ifc', 'nosym', 'loto_2d',
                                    'loto_disable']
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)



import os
from ase.calculators.calculator import FileIOCalculator, CalculationFailed
from xespresso.xio import read_espresso_asei, write_espresso_asei
import copy
import logging
logger = logging.getLogger(__name__)


class PostCalculation:
    
    package = 'dos'
    package_parameters = {}

    def __init__(self, parent_directory, prefix, queue = False, parallel = '', **kwargs) -> None:
        self.parent_directory = parent_directory
        self.prefix = prefix
        self.queue = queue
        self.parallel = parallel
        self.parameters = kwargs
        self.directory = os.path.join(self.parent_directory, '%s/'%self.package)
        self.set_label(self.directory, self.prefix)
        self.parameters['prefix'] = self.prefix
        self.parameters['outdir'] = '../'
        self.state_info = None

    def set_label(self, label, prefix):
        '''
        set directory and prefix from label
        '''
        self.directory = label
        if not prefix:
            self.prefix = os.path.split(label)[1]
        else:
            self.prefix = prefix
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        self.label = os.path.join(self.directory, self.prefix)
        self.asei = os.path.join(self.directory, '%s.asei'%self.prefix)
        self.asei_temp = os.path.join(self.directory, '.%s.asei_temp'%self.prefix)
        self.post_asei = os.path.join(self.directory, '%s.post_asei'%self.prefix)
        self.save_directory = os.path.join(self.directory, '%s.save'%self.prefix)
        logger.debug('Directory: %s'%(self.directory))
        logger.debug('Prefix: %s'%(self.prefix))

    def run(self):
        '''
        todo: 
        '''
        from xespresso.scheduler import set_queue
        
        print('{0:=^60}'.format(self.package))
        if self.check_state() == 0:
            return
        self.write_input()
        set_queue(self, package = self.package, parallel = self.parallel, queue = self.queue)
        self.post_calculate()
        self.post_read_results()

    def check_state(self):
        from xespresso.utils import get_hash
        self.state_info = None
        for wfc in ['wfc1', 'wfcdw1']:
            wfcFile = os.path.join(self.parent_directory, '%s.save/%s.dat'%(self.prefix, wfc))
            if os.path.isfile(wfcFile):
                self.state_info = get_hash(wfcFile)
        output, meg = self.read_convergence_post(self.package)
        if output:
            logger.debug('Previous calculation done.')
            if os.path.isfile(self.post_asei):
                system_changes =  self.check_state_post(self.post_asei, self.package)
                if not system_changes:
                    logger.debug('File and Parameters did not change. Use previous results!')
                    return 0
        return 1

    def check_state_post(self, asei, package):
        old_state_info, old_parameters = read_espresso_asei(asei, package)
        if not self.state_info == old_state_info:
            logger.debug('File in save changed')
            return True
        elif not self.parameters == old_parameters:
            logger.debug('Parameters changed')
            return True
        else:
            return False

    def write_input(self):
        self.write_package_input()
        write_espresso_asei(self.post_asei, self.state_info, self.parameters)

    def write_package_input(self):
        filename = os.path.join(self.directory, '%s.%si' %(self.prefix, self.package))
        with open(filename, 'w') as f:
            for section, parameters in self.package_parameters.items():
                # logger.debug(section)
                if section != 'LINE':
                    f.write('&%s\n'%section)
                    for key, value in self.parameters.items():
                        if key in parameters:
                            if isinstance(value, dict):
                                for subkey, subvalue in value.items():
                                    if isinstance(subvalue, str):
                                        f.write('  %s(%s) = "%s", \n' %(key, subkey, subvalue))
                                    else:
                                        f.write('  %s(%s) = %s, \n' %(key, subkey, subvalue))
                            else:
                                if isinstance(value, str):
                                    f.write('  {0:10s} =  "{1}" \n'.format(key, value))
                                else:
                                    f.write('  {0:10s} =  {1} \n'.format(key, value))
                    f.write('/ \n')
                else:
                    for key, value in self.parameters.items():
                        if key in parameters:
                            f.write('  %s \n' %(value))


    def post_calculate(self):
        import subprocess
        
        command = self.command
        print('Running %s'%self.package)
        try:
            proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        except OSError as err:
            msg = 'Failed to execute "{}"'.format(command)
            raise EnvironmentError(msg) from err

        errorcode = proc.wait()

        if errorcode:
            path = os.path.abspath(self.directory)
            msg = ('Command "{}" failed in '
                   '{} with error code {}'.format(command,
                                                  path, errorcode))
            raise CalculationFailed(msg)
        print('Done: %s' %self.package)
    
    def post_read_results(self):
        '''
        '''
        pass

    def read_convergence_post(self, package = 'pw'):
        '''
        Read the status of the calculation.
        {
        '0': 'Done',
        }
        '''
        
        output = self.label + '.%so'%package
        if not os.path.exists(output):
            # print('%s not exists'%output)
            return False, 'No pwo output file'
        with open(output, 'r') as f:
            lines = f.readlines()
            if len(lines) == 0: return False, 'pwo file has nothing'
            nlines = len(lines)
            n = min([100, nlines])
            for line in lines[-n:-1]:
                if line.rfind('JOB DONE.') > -1:
                    logger.debug('JOB DONE.')
                    return True, line
        return False, line
    



from xespresso.post.base import PostCalculation


class EspressoDos(PostCalculation):

    package = 'dos'
    package_parameters = {'DOS': ['prefix', 'outdir', 'bz_sum', 'ngauss', 'degauss',
                                  'Emin', 'Emax', 'DeltaE', 'fildo', ]
                          }

    def __init__(self, parent_directory, prefix, queue=False, parallel='', **kwargs) -> None:
        super().__init__(parent_directory, prefix, queue=False, parallel='', **kwargs)



'''
0 bulk
1 clean surfaces, different termination
2 add OH* and O*, different coverage
3 build Pourbaix diagram
4 choose surface 
5 different OER site
6 add O*, OH* and OOH*
7 free energy diagram
8 report

'''
import os
import numpy as np
from copy import deepcopy
from ase.atoms import Atoms
from ase.io import write
from pymatgen.analysis.adsorption import AdsorbateSiteFinder
from pymatgen.core.surface import SlabGenerator
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.ase import AseAtomsAdaptor
from xespresso import Espresso
from xespresso.tools import mypool, fix_layers, dipole_correction
from xespresso.workflow.base import Base
from xespresso.xlog import XLogger

import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor

# ----------------------------------


# view(mols.values())
class OER_base(Base):
    def __init__(self, atoms, name='oer', label='.', prefix=None, calculator=None, molecule_energies=None, view=False, debug=False):
        Base.__init__(self, atoms, label=label, prefix=prefix,
                      calculator=calculator, view=view, debug=debug)
        self.name = name
        self.set_logger(OERLogger)
        self.molecule_energies = molecule_energies
        self.mols = {
            'O': Atoms('O'),
            'OH': Atoms('OH', positions=[[0, 0, 0], [0.6, 0.7, 0.15]]),
            'OOH': Atoms('OOH', positions=[[0, 0, 0], [0.8, 0.9, 0.0], [1.4, 1.6, 0.45]]),
        }
        '''
        from qedatas import E_h2o, E_h2, G_o2
        E_h2o = E_h2o + 0.56 - 0.67
        E_h2 = E_h2 + 0.27 - 0.41
        G_o2 = 0.10 - 0.64
        '''

    def get_sites_dict(self, atoms, excludes=[]):
        '''
        get the surface site for adsorption
        1) for OER_pourbaix, ['O', 'N'] sites are generally excluded.
        2) for OER_site
                metal: Ti, La -> OH -> O -> OOH -> O2
                O:          O -> OOH -> O2 -> OH 
                H:          OH -> O -> OOH -> O2

        Return
        sites: dict
            e.g. {'Pt': [0.0, 0.0, 5.0]}
        '''
        from scipy.spatial.distance import squareform, pdist
        slabs = AseAtomsAdaptor.get_structure(atoms)
        asf_slabs = AdsorbateSiteFinder(slabs)
        ads_sites = asf_slabs.find_adsorption_sites(distance=0.0)
        sites = {}
        count = 0
        # view(atoms)
        for pos in ads_sites['ontop']:
            poss = atoms.get_positions()
            poss = np.append(poss, pos).reshape(-1, 3)
            ind = np.argsort(squareform(pdist(poss))[-1])[1] - 1
            # print(ind)
            symbol = atoms[ind].symbol
            if symbol in excludes:
                continue
            sites['site-%s-%s' % (count, symbol)] = pos + np.array([0, 0, 2.0])
        return sites

    def vib_zpe(self, job, atoms):
        from ase.vibrations import Vibrations
        self.log('-'*60)
        self.log('Run ZEP calculation: {0}'.format(job))
        label = os.path.join(self.label, job)
        label = os.path.join(label, 'vib')
        # copy file
        calculator = deepcopy(self.calculator)
        dip = dipole_correction(atoms, edir=3)
        calculator.update(dip)
        calculator.update({'calculation': 'scf',
                           'tstress': True,
                           'tprnfor': True,
                           'outdir': '../',
                           'prefix': '%s' % job,
                           'startingpot': 'file',
                           'startingwfc': 'file',
                           'etot_conv_thr': 1e-6,
                           'disk_io': 'none',
                           })
        # better to restart from previous geo_relax calculation
        calc = Espresso(label=label,
                        **calculator,
                        )
        atoms.calc = calc
        if job[-1] == 'O':
            indices = [-1]
        elif job[-2:] == 'OH' and job[-3:] != 'OOH':
            indices = [-1, -2]
        elif job[-3:] == 'OOH':
            indices = [-1, -2, -3]
        elif job[-2:] == 'O2':
            indices = [-1, -2]
        else:
            indices = []
            # print('%!!!')
            return job, 0
        vib = Vibrations(atoms, name=os.path.join(label, job), indices=indices)
        vib.run()
        vib_energies = np.real(vib.get_energies())
        zpe = 0.
        for energy in vib_energies:
            zpe += 0.5 * energy
        self.zpes[job] = zpe
        return job, zpe

    def pdos(self, job, atoms):
        """
        calculate the pdos
        """
        self.log('-'*60)
        self.log('Run pdos calculation: {0}'.format(job))
        calc = self.calcs[job]
        fe = calc.get_fermi_level()
        atoms = calc.results['atoms']
        calculator = deepcopy(self.calculator)
        kpts = [calculator['kpts'][0]*2, calculator['kpts']
                [1]*2, calculator['kpts'][2]]
        if 'parallel' not in calculator:
            calculator['parallel'] = '-nk 1'
        calc.nscf(queue=calculator['queue'], parallel=calculator['parallel'],
                  occupations='tetrahedra', kpts=kpts)
        calc.nscf_calculate()
        calc.read_results()
        calc.post(queue=calculator['queue'], package='dos',
                  Emin=fe - 30, Emax=fe + 30, DeltaE=0.01)
        calc.post(queue=calculator['queue'], package='projwfc',
                  Emin=fe - 30, Emax=fe + 30, DeltaE=0.01)


class OER_bulk(OER_base):
    def __init__(self, atoms, label='.', prefix=None, surfaces={}, indexs=[], min_slab_size=6.0, min_vacuum_size=15.0,  nlayer=4, fix=[0, 3], tol=1.0, termination=None, pourbaix_input={}, surface_input={}, calculator=None, molecule_energies=None, view=False, debug=False):
        '''
        Generate surface slab model from bulk structure

        To do: adding vc-relax for the bulk
        '''
        OER_base.__init__(self, atoms, label=label, prefix=prefix,
                          calculator=calculator, view=view, molecule_energies=molecule_energies)
        self.name = 'bulk'
        self.children = {}
        self.surfaces = surfaces
        self.indexs = indexs
        self.min_slab_size = min_slab_size
        self.min_vacuum_size = min_vacuum_size
        self.nlayer = nlayer
        self.fix = fix
        self.tol = tol
        self.pourbaix_input = pourbaix_input
        self.surface_input = surface_input

    def get_slabs(self, atoms=None, index=None):
        """
        """
        from math import ceil
        if not atoms:
            atoms = self.atoms
        struct = AseAtomsAdaptor.get_structure(self.atoms)
        struct = SpacegroupAnalyzer(
            struct).get_conventional_standard_structure()
        slabgen = SlabGenerator(struct,
                                miller_index=index,
                                min_slab_size=self.min_slab_size,
                                min_vacuum_size=self.min_vacuum_size,
                                center_slab=False)
        slabs = {}
        for n, slab in enumerate(slabgen.get_slabs()):
            slab = AseAtomsAdaptor.get_atoms(slab)
            self.set_ase_cell(slab)
            a, b, c, alpha, beta, gamma = slab.get_cell_lengths_and_angles()
            supercell = [ceil(5/a), ceil(5/b), ceil(5/c)]
            # print(a, b, c, supercell)
            slab = slab*supercell
            slabs[n] = slab.copy()
        return slabs

    def set_ase_cell(self, surf):
        from numpy.linalg import norm
        a1, a2, a3 = surf.cell
        surf.set_cell([a1, a2,
                       np.cross(a1, a2) * np.dot(a3, np.cross(a1, a2)) /
                       norm(np.cross(a1, a2))**2])
        #
        a1, a2, a3 = surf.cell
        surf.set_cell([(norm(a1), 0, 0),
                       (np.dot(a1, a2) / norm(a1),
                        np.sqrt(norm(a2)**2 - (np.dot(a1, a2) / norm(a1))**2), 0),
                       (0, 0, norm(a3))],
                      scale_atoms=True)
        surf.wrap()

    def build_children(self, surfaces=None, nlayer=None, fix=None, tol=None):
        '''
        surfaces and terminations

        '''
        if not surfaces:
            surfaces = self.surfaces
        if not nlayer:
            nlayer = self.nlayer
        if not fix:
            fix = self.fix
        if not tol:
            tol = self.tol
        self.log('-'*60)
        self.log('Build surface:\n')
        count = 0
        for index, ters in self.indexs.items():
            count += 1
            self.log('{0:15s}: {1}'.format('%s. index' % count, index))
            # atoms = surface(self.atoms, index, nlayer + 2, tol = self.tol, vacuum=1)
            # atoms.pbc = [True, True, True]
            # print(self.tol)
            # layers = get_layers(atoms, (0, 0, 1), self.tol)[0]
            # print(layers)
            # terminations = self.get_terminations(atoms, layers)
            terminations = self.get_slabs(self.atoms, index)
            print(terminations)
            # if len(ters) == 0:
            # ters = terminations.keys()
            for ter, surf in terminations.items():
                # if ter not in ters: continue
                self.log('{0:15s}: {1}'.format('    termination', ter))
                # surf = atoms.copy()
                # index = [j for j in range(len(surf)) if layers[j] in range(ind, ind - nlayer, -1)]
                # surf = surf[index]
                # surf.positions[:, 2] -= min(surf.positions[:, 2]) - 0.01
                # surf.cell[2][2] = max(surf.positions[:, 2]) + 15
                surf = fix_layers(surf, (0, 0, 1), tol, fix)
                #
                prefix = '%s%s%s-%s' % (index[0], index[1], index[2], ter)
                self.images[prefix] = surf
                surf = OER_pourbaix(surf, label=os.path.join(self.label, prefix), prefix=prefix, calculator=self.calculator,
                                    view=self.view, molecule_energies=self.molecule_energies, **self.pourbaix_input, surface_input=self.surface_input)
                self.children[prefix] = surf
            self.log()
        self.log('Total number of surfaces: {0}\n'.format(len(self.children)))

    def get_terminations(self, slab, layers):
        '''
        determine how many terminations for a surface index
        for a slab with SrO, TiO2, SrO, TiO2 ... ordering, there will be two terminations
        'SrO' and 'TiO2' 
        Search from the top to the bottom
        '''
        from ase.build import minimize_rotation_and_translation
        from ase.calculators.calculator import compare_atoms, equal
        import pprint
        nlayer = max(layers) + 1
        terminations = {}
        natoms = len(slab)
        for i in range(nlayer - 2, 0, -1):
            atoms = slab[layers == i]
            formula = atoms.get_chemical_formula(mode='hill')
            print(formula)
            inds = [j for j in range(natoms) if layers[j]
                    in range(i, i - self.nlayer, -1)]
            atoms = slab[inds]
            atoms.wrap()
            if not terminations:
                terminations['%s-%s' % (formula, i)] = atoms.copy()
            else:
                flag = True
                atoms1 = atoms.copy()
                for ter, atoms2 in terminations.items():
                    if len(atoms1) != len(atoms2):
                        flag = False
                    else:
                        minimize_rotation_and_translation(atoms2, atoms1)
                        if equal(atoms1.arrays, atoms2.arrays, tol=0.1):
                            flag = False
                if flag:
                    terminations['%s-%s' % (formula, i)] = atoms.copy()
                    print(formula, ter)
                    # view([atoms1, atoms2])
        pprint.pprint(terminations)
        # view(terminations.values())
        return terminations


class OER_pourbaix(OER_base):
    def __init__(self, atoms, label='', prefix=None, sites_dict=None, calculator=None, coverages=[0, 1], adsorbates=['O', 'OH'], surface_input={}, molecule_energies=None, view=False, debug=False):
        '''
        Calculate surface Pourbaix diagram.
        coverages: list
            A list of float, e.g. [0, 0.5, 1]
        adsorbates: list
            The adsorbate used to build the coverage, support ['O', 'OH']
        sites_dict: dict
            In OER_pourbaix, site is position. e.g. {'fcc': [0.5, 0.5, 3]}. 
            If None, the ontop site will be found and used.
        '''
        self.name = 'pourbaix'
        OER_base.__init__(self, atoms, label=label, prefix=prefix,
                          calculator=calculator, view=view, molecule_energies=molecule_energies)
        if not sites_dict:
            self.sites_dict = self.get_sites_dict(
                self.atoms, excludes=['O', 'N'])
        else:
            self.sites_dict = sites_dict
        self.coverages = coverages
        self.adsorbates = [''] + adsorbates
        self.surface_input = surface_input
        self.images = {}
        self.results = {}
        self.children = {'0-O-0-OH': atoms}

    def run(self):
        # coverage
        self.build_coverage()
        if self.view:
            self.view_images()
            return
        #
        self.run_coverage()
        # fig, ax = plt.subplots(1, 1)
        ax, children = self.plot_pourbaix_diagram(self.results)  # , ax = ax)
        filename = os.path.join(
            self.label, '%s-pourbaix_diagram.png' % self.prefix)
        # fig.savefig(filename)
        self.log('-'*60)
        self.log('Save pourbaix diagram to: {0}'.format(filename))
        self.log('Stable surfaces from pourbaix diagram:')
        for prefix in children:
            surf = self.images[prefix]
            self.images[prefix] = surf
            surf = OER_surface(surf, label=os.path.join(self.label, prefix), prefix=prefix,
                               calculator=self.calculator, molecule_energies=self.molecule_energies, **self.surface_input)
            self.log('{0:15s}: {1}'.format('    Stable surface', prefix))
            self.children[prefix] = surf
        self.log()
        self.log('Total number of stable surfaces: {0}\n'.format(
            len(self.children)))
        # OER
        self.run_children()

    def run_coverage(self):
        self.pool_atoms(self.images)
        self.log('-'*60)
        self.log('Image energies (eV):')
        for job, result in self.results.items():
            self.log('{0:10s}  Energy: {1:1.3f}'.format(job, result['energy']))

    def build_coverage(self, coverages=None, sites_dict=None):
        '''
        add 1ML O* and OH*
        could be:
        1 ML O*
        1 ML OH*
        0.5 ML O* and 0.5 ML OH*
        '''
        import itertools
        if coverages is None:
            coverages = self.coverages
        if sites_dict is None:
            sites_dict = self.sites_dict
        self.sites = []
        site_mols = []
        self.log('-'*60)
        self.log('Build coverages: ')
        self.log('Surface sites: ')
        self.log('    {0:10s} {1:4s}'.format('Symbol', 'Position'))
        count = 0
        for symbol, position in sites_dict.items():
            # self.log('    %s %s'%(symbol, len(positions)))
            count += 1
            self.log('    {0:3d}   {1:10s} {2}'.format(
                count, symbol, position))
            self.sites.append(position)
            site_mols.append(self.adsorbates)
        self.nsite = len(self.sites)
        self.log('Number of sites: {0}'.format(self.nsite))
        if len(sites_dict) == 0:
            return
        combinations = list(itertools.product(*site_mols))
        self.log('Coverages: {0}'.format(self.coverages))
        for combination in combinations:
            noh = combination.count('OH')
            no = combination.count('O')
            coverage = (no + noh)/self.nsite
            flag = [abs(coverage - x) for x in coverages]
            if min(flag) > 0.1:
                continue
            formula = '%s-O-%s-OH' % (no, noh)
            if formula in self.images:
                continue
            ads = Atoms('')
            for i in range(len(combination)):
                if not combination[i]:
                    continue
                pos = self.sites[i]
                mol = self.mols[combination[i]].copy()
                mol.translate(pos - mol[0].position + [0, 0, 1.5])
                ads = ads + mol
            natoms = self.atoms.copy()
            self.images[formula] = natoms + ads
            self.log('Structure: {0}'.format(formula))
        self.log('Total number of structures: {0}'.format(len(self.images)))

    def plot_pourbaix_diagram(self, results, ax=None, G_O=None, G_OH=None):
        '''
        mh = potential
        '''
        E_h2o = self.molecule_energies['H2O']
        E_h2 = self.molecule_energies['H2']
        if not G_O:
            G_O = E_h2o - E_h2
        if not G_OH:
            G_OH = E_h2o - E_h2/2.0
        potential = np.linspace(0, 2, 20)
        # if ax is None:
        # ax = plt.gca()
        E0 = results['0-O-0-OH']['energy']
        for job, result in results.items():
            E = result['energy']
            no = int(job.split('-')[0])
            noh = int(job.split('-')[2])
            dE = E - E0 - noh*(G_OH + potential) - no*(G_O + 2*potential)
            # ax.plot(potential, dE, label = job)
            self.results[job]['pourbaix_energy'] = dE
        # ax.legend()
        # ax.set_xlabel('Potential (V)')
        # ax.set_ylabel('$\Delta$ G (eV)')
        StableSurfaces = []
        for i in range(20):
            Emin = 1000
            jobmin = ''
            for job, result in self.results.items():
                E = result['pourbaix_energy']
                if E[i] < Emin:
                    Emin = E[i]
                    jobmin = job
            StableSurfaces.append(jobmin)
        StableSurfaces = list(set(StableSurfaces))
        return ax, StableSurfaces


class OER_surface(OER_base):
    def __init__(self, atoms, label='', prefix=None, activate_species=None, sites_dict=None, calculator=None, molecule_energies=None):
        '''
        sites_dict: dict
            the active site, should be the atom in the structure. e.g. {'Pt': [2]}. 
            If None, the ontop atom will be found and used.
            This is different from the site in OER_pourbaix.
        '''
        self.name = 'surface'
        OER_base.__init__(self, atoms, label=label, prefix=prefix,
                          calculator=calculator, molecule_energies=molecule_energies)
        self.atoms = atoms
        self.activate_species = activate_species
        if not sites_dict:
            self.sites_dict = self.get_sites_dict(self.atoms)
        else:
            self.sites_dict = sites_dict
        self.children = {}

    def run(self):
        self.build_children()
        self.run_children()
        return 0

    def build_children(self, ):
        '''
        '''
        self.log('-'*60)
        self.log('OER sites:')
        for prefix, pos in self.sites_dict.items():
            ele = prefix.split('-')[-1]
            if ele not in self.activate_species:
                continue
            self.images[prefix] = self.atoms
            child = OER_site(self.atoms, label=os.path.join(self.label, prefix), position=pos,
                             prefix=prefix, calculator=self.calculator, molecule_energies=self.molecule_energies)
            self.log('{0:15s}: {1}'.format('    sites', prefix))
            self.children[prefix] = child
            self.log()
        self.log('Total number of OER sites: {0}\n'.format(len(self.children)))


class OER_site(OER_base):
    def __init__(self, atoms, label='', prefix=None, site_type='ontop', site=None, height=2.0, calculator=None, molecule_energies=None, mols=None, view=False, debug=False):
        '''
        Workflow:
        0. Read the surface slab
        1. Build the adsorption site
        2. Add O*, OH* and OOH*, and relax the structure
        3. Calculate the dos and pdos
        4. Calculate the ZPE energy using Vibrations module
        5. Calculate the Gibbs free energy
        6. Generate the report for OER overpotential

        atoms:
            The surface slab, from previous optmization calculation.
        model_type: 
            (1) 'ontop', (2) 'bridge', (3) 'hollow'. For 'ontop' site, the 'O' and 'H' vacancy pathway are possible.
        site: 
            index for the adsoprtion. 
        height: 
            height of the O*, OH* and OOH* from the adsorption site.

        Examples:
        >>> oer = OER_site(atoms_opt,
                    label = 'oer/Pt-001-ontop',
                    site_type = 'ontop',
                    site = -1,
                    height=2.0,
                        calculator = parameters, 
                    molecule_energies = molecule_energies,
                    )
        >>> oer.run()

        '''
        self.name = 'site'
        OER_base.__init__(self, atoms, label=label, prefix=prefix, calculator=calculator,
                          molecule_energies=molecule_energies, view=view, debug=debug)
        self.images = {}
        self.site_type = site_type
        self.site = site
        self.height = height
        self.site_symbol = site_type
        if site_type.upper() == 'ONTOP':
            self.site_symbol = atoms[site].symbol
            self.site_position = atoms[site].position + \
                np.array([0, 0, height])
        elif site_type.upper() == 'BRIDGE':
            self.site_position = (
                atoms[site[0]].position + atoms[site[1]].position)/2.0 + np.array([0, 0, height])
        elif site_type.upper() == 'HOLLOW':
            self.site_position = (atoms[site[0]].position +
                                  atoms[site[1]].position +
                                  atoms[site[2]].position)/2.0 + np.array([0, 0, height])
        elif site_type.upper() == 'POSITION':
            self.site_position = site + np.array([0, 0, height])
        if mols:
            self.mols = mols
        self.children = {}
        self.calcs = {}
        self.zpes = {}
        self.free_energies = {}

    def run(self):
        """
        Function build all the main workflow
        """
        #
        self.build_oer_adsorbate()
        if self.view:
            self.view_images(self.children)
            return
        # geometry relaxiton
        self.pool_atoms(self.children, self.geo_relax_espresso)
        self.log('-'*60)
        self.log('Image energies (eV):')
        # self.log('{0: <20}  Energy: {1:1.3f}'.format('Clean Surface', self.atoms.calc.results['energy']))
        for job, result in self.results.items():
            self.log('    {0:<20}  {1:1.3f}'.format(job, result['energy']))
            self.children[job] = result['atoms']
        # pdos
        self.pool_atoms(self.children, self.pdos)
        # vibration calculation to get zero point energy
        self.pool_atoms(self.children, self.vib_zpe)
        self.log('-'*60)
        self.log('ZPE energies (eV):')
        for job, zpe in self.zpes.items():
            self.log('    {0:<20}  {1:1.3f}'.format(job, zpe))
        # calculate the oer over potential
        self.get_free_energy()
        self.write_image()
        self.log('-'*60)
        self.log('Gibbs free energies (eV): ')
        for step, e in zip(self.steps, self.free_energies):
            self.log('    {0:<10}    {1:1.2f}'.format(step, e))
        self.log('{0:10s}: {1:1.2f} eV'.format(
            'Over potential', self.over_potential))
        print('{0:10s}: {1:1.2f} eV'.format(
            'Over potential', self.over_potential))
        # view(self.children.values())
        return 'Over potential', self.over_potential

    def build_oer_adsorbate(self, ):
        """
        Add O*, OH* and OOH*, and relax the structure
        For 'ontop' site, the 'O' and 'H' vacancy pathway are possible.
        """
        self.log('-'*60)
        self.log('Adsoption site: index ({0})  symbol({1})'.format(
            self.site, self.site_symbol))
        self.log('OER adsorbate:')
        atoms = self.atoms
        label = self.label.replace('/', '_')
        # print(self.site_position)
        job = '%s_%s' % (label, 'Clean')
        self.children[job] = self.atoms.copy()
        if self.site_symbol not in ['H', 'O']:
            for prefix, mol in self.mols.items():
                # print(prefix, mol)
                job = '%s_%s' % (label, prefix)
                self.log('    %s' % job)
                ads = mol.copy()
                natoms = atoms.copy()
                ads.translate(self.site_position - ads[0].position)
                natoms = natoms + ads
                self.children[job] = natoms
        # O
        ind = self.site
        if self.site_symbol == 'O':
            job = '%s_%s' % (label, 'O')
            self.log('    %s' % job)
            self.children[job] = self.atoms.copy()
            atoms = self.atoms.copy()
            del atoms[[ind]]
            job = '%s_%s' % (label, 'Clean')
            self.log('    %s' % job)
            self.children[job] = atoms.copy()
            for prefix, mol in self.mols.items():
                # print(prefix, mol)
                if prefix == 'O':
                    continue
                job = '%s_%s' % (label, prefix)
                self.log('    %s' % job)
                ads = mol.copy()
                natoms = atoms.copy()
                ads.translate(self.atoms[ind].position - ads[0].position)
                natoms = natoms + ads
                self.children[job] = natoms
        elif self.site_symbol == 'H':
            atoms = self.atoms.copy()
            dis = atoms.get_distance(ind, range(len(atoms)))
            indo = dis.index(min(dis))
            # O
            job = '%s_%s' % (label, 'O')
            self.log('    %s' % job)
            natoms = atoms.copy()
            del natoms[[ind]]
            self.children[job] = natoms
            # O2
            job = '%s_%s' % (label, 'Clean')
            self.log('    %s' % job)
            natoms = atoms.copy()
            del natoms[[ind, indo]]
            self.children[job] = natoms
            # OOH
            job = '%s_%s' % (label, 'OOH')
            self.log('    %s' % job)
            ads = self.mols[job].copy()
            natoms = atoms.copy()
            ads.translate(self.atoms[indo].position - ads[0].position)
            natoms = natoms + ads
            self.children[job] = natoms
        self.log('Total number of OER adsorbate: {0}\n'.format(
            len(self.children)))

    def get_free_energy(self, results=None, G_O=None, G_OH=None, G_OOH=None, E_H=None):
        """
        """
        G_h2o = self.molecule_energies['H2O']
        G_h2 = self.molecule_energies['H2']
        if not G_O:
            G_O = G_h2o - G_h2
        if not G_OH:
            G_OH = G_h2o - G_h2/2.0
        if not G_OOH:
            G_OOH = 2*G_h2o - 3*G_h2/2.0
        if not E_H:
            E_H = G_h2/2.0
        # fig, ax = plt.subplots()
        label = self.label.replace('/', '_')
        E0 = self.results['%s_%s' % (label, 'Clean')]['energy']
        # if self.atoms[self.site].symbol not in []:
        dG_OH = self.results['%s_%s' % (
            label, 'OH')]['energy'] - E0 - G_OH + self.zpes['%s_%s' % (label, 'OH')]
        dG_O = self.results['%s_%s' % (
            label, 'O')]['energy'] - E0 - G_O + self.zpes['%s_%s' % (label, 'O')]
        dG_OOH = self.results['%s_%s' % (
            label, 'OOH')]['energy'] - E0 - G_OOH + self.zpes['%s_%s' % (label, 'OOH')]
        dG_O2 = 4.92
        self.steps = ['OH', 'O', 'OOH', 'O2']
        self.free_energies = [dG_OH, dG_O, dG_OOH, dG_O2]
        # self.plot_free_energy([0, G_OH, G_O, G_OOH, G_O2], ['*', 'OH', 'O', 'OOH', 'O2'], ax = ax)
        # if self.atoms[self.site].symbol == 'O':
        #     dG_OOH = self.results['%s_%s'%(label, 'OOH')]['energy'] - E0 - G_OH + self.zpes['%s_%s'%(label, 'OOH')]
        #     dG_O2 = self.results['%s_%s'%(label, 'O2')]['energy'] - E0 + G_O + self.zpes['%s_%s'%(label, 'O2')]
        #     dG_OH = self.results['%s_%s'%(label, 'OH')]['energy'] - E0 - E_H + self.zpes['%s_%s'%(label, 'OH')]
        #     dG_O = 4.92
        #     self.steps = ['OOH', 'O2', 'OH', 'O']
        #     self.free_energies = [dG_OOH, dG_O2, dG_OH, dG_O]
        #     # self.plot_free_energy([0, G_OOH, G_O2, G_OH, G_O], ['O', 'OOH', 'O2', 'OH', 'O'], ax = ax)
        # if self.atoms[self.site].symbol == 'H':
        #     dG_O = self.results['%s_%s'%(label, 'OH')]['energy'] - E0 + E_H + self.zpes['%s_%s'%(label, 'OH')]
        #     dG_OOH = self.results['%s_%s'%(label, 'O')]['energy'] - E0 - G_O + self.zpes['%s_%s'%(label, 'O')]
        #     dG_O2 = self.results['%s_%s'%(label, 'O2')]['energy'] - E0 + G_OH + self.zpes['%s_%s'%(label, 'O2')]
        #     dG_OH = 4.92
        #     self.steps = ['O', 'OOH', '2', 'OH']
        #     self.free_energies = [dG_O, dG_OOH, dG_O2, dG_OH]
        #     # self.plot_free_energy([0, G_O, G_OOH, G_O2, G_OH], ['OH', 'O', 'OOH', 'O2', 'OH'], ax = ax)
        # fig.savefig(os.path.join(self.label, '%s-free-energy.png'%self.prefix))
        self.over_potential = max([self.free_energies[1] - self.free_energies[0],
                                   self.free_energies[2] -
                                   self.free_energies[1],
                                   self.free_energies[3] - self.free_energies[2]]) - 1.23

    def write_image(self, ):
        for job in self.results:
            atoms = self.results[job]['atoms']
            write(os.path.join(self.label, '%s-top.png' %
                               job), atoms * (2, 2, 1))
            write(os.path.join(self.label, '%s-side.png' %
                               job), atoms * (2, 2, 1), rotation='-90x')
            write(os.path.join(self.label, '%s.png' % job),
                  atoms * (2, 2, 1), rotation='10z,-80x')

    def plot_free_energy(self, E, labels, ax=None):
        if ax is None:
            ax = plt.gca()
        ax.plot([0, 1], [0, 0], '-')
        ax.text(0 + 0.3, E[0] + 0.5, labels[0])
        for i in range(1, 5):
            ax.plot([i, i], [E[i - 1], E[i]], '-')
            ax.plot([i, i+1], [E[i], E[i]], '-')
            ax.text(i + 0.3, E[i] + 0.5, labels[i])
        ax.set_xlabel('Step')
        ax.set_ylabel('$\Delta$ G (eV)')
        return ax


class OERLogger(XLogger):
    """Class for handling all text output."""

    def __init__(self, ):
        XLogger.__init__(self,)

    def logo(self):
        self()
        self('  //==\\   //====  //====\\  ')
        self(' ||    ||  ||      ||    || ')
        self(' ||    ||  ======  |\====// ')
        self(' ||    ||  ||      || \\    ')
        self('  \\==//   \\====  ||  \\   ')
        self()
